#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import socket
import unittest
import unittest.mock

import PIL.Image
import requests

import sacad


sacad.web_cache.DISABLE_PERSISTENT_CACHING = True


def is_internet_reachable():
  try:
    # open TCP socket to Google DNS server
    socket.create_connection(("8.8.8.8", 53))
  except OSError as e:
    if e.errno == 101:
      return False
    raise
  return True


def download(url, filepath):
  response = requests.get(url, timeout=5, verify=False, stream=True)
  response.raise_for_status()
  with open(filepath, "wb") as f:
    for chunk in response.iter_content(2 ** 14):
      f.write(chunk)


@unittest.skipUnless(is_internet_reachable(), "Need Internet access")
class TestSacad(unittest.TestCase):

  @staticmethod
  def getImgInfo(img_filepath):
    with open(img_filepath, "rb") as img_file:
      img = PIL.Image.open(img_file)
      format = img.format.lower()
      format = sacad.SUPPORTED_IMG_FORMATS[format]
      width, height = img.size
    return format, width, height

  def test_getMasterOfPuppetsCover(self):
    """ Search and download cover for 'Master of Puppets' with different parameters. """
    for format in sacad.cover.CoverImageFormat:
      for size in (300, 600, 1200):
        for size_tolerance in (0, 25, 50):
          for prefer_https in (True, False):
            with sacad.mkstemp_ctx.mkstemp(prefix="sacad_test_",
                                           suffix=".%s" % (format.name.lower())) as tmp_filepath:
              sacad.main("Master of Puppets", "Metallica", format, size, size_tolerance, False, prefer_https, tmp_filepath)
              out_format, out_width, out_height = __class__.getImgInfo(tmp_filepath)
              self.assertEqual(out_format, format)
              self.assertLessEqual(out_width, size * (100 + size_tolerance) / 100)
              self.assertGreaterEqual(out_width, size * (100 - size_tolerance) / 100)
              self.assertLessEqual(out_height, size * (100 + size_tolerance) / 100)
              self.assertGreaterEqual(out_height, size * (100 - size_tolerance) / 100)

  def test_getImageUrlMetadata(self):
    """ Download the beginning of an image file to guess its format and resolution. """
    url = "http://lacuriosphere.fr/wp-content/uploads/2013/12/mountains-and-clouds.jpg"
    sacad.CoverSourceResult.getImageMetadata = unittest.mock.Mock(wraps=sacad.CoverSourceResult.getImageMetadata)
    cover = sacad.CoverSourceResult(url, None, None, thumbnail_url=None, source_quality=None)
    cover.updateImageMetadata()
    self.assertEqual(cover.size, (1600, 1200))
    self.assertEqual(cover.format, sacad.cover.CoverImageFormat.JPEG)
    self.assertGreaterEqual(sacad.CoverSourceResult.getImageMetadata.call_count, 0)
    self.assertLessEqual(sacad.CoverSourceResult.getImageMetadata.call_count, 3)

  def test_compareImageSignatures(self):
    """ Compare images using their signatures. """
    urls = ("http://wac.450f.edgecastcdn.net/80450F/kool1017.com/files/2013/09/cover_highway_to_hell_500x500.jpg",
            "http://www.jesus-is-savior.com/Evils%20in%20America/Rock-n-Roll/highway_to_hell-large.jpg",
            "http://i158.photobucket.com/albums/t113/gatershanks/Glee%20Alternative%20Song%20Covers/1x14%20Hell%20O/1x14Hell-O-HighwayToHell.jpg")
    with sacad.mkstemp_ctx.mkstemp(suffix="jpg") as temp_filepath1, \
            sacad.mkstemp_ctx.mkstemp(suffix="jpg") as temp_filepath2, \
            sacad.mkstemp_ctx.mkstemp(suffix="jpg") as temp_filepath3:
      img_sig = {}
      for i, (url, filepath) in enumerate(zip(urls, (temp_filepath1, temp_filepath2, temp_filepath3))):
        download(url, filepath)
        with open(filepath, "rb") as img_file:
          img_data = img_file.read()
          img_sig[i] = sacad.CoverSourceResult.computeImgSignature(img_data)
      self.assertTrue(sacad.CoverSourceResult.areImageSigsSimilar(img_sig[0], img_sig[1]))
      self.assertTrue(sacad.CoverSourceResult.areImageSigsSimilar(img_sig[1], img_sig[0]))
      self.assertFalse(sacad.CoverSourceResult.areImageSigsSimilar(img_sig[0], img_sig[2]))
      self.assertFalse(sacad.CoverSourceResult.areImageSigsSimilar(img_sig[1], img_sig[2]))
      self.assertFalse(sacad.CoverSourceResult.areImageSigsSimilar(img_sig[2], img_sig[0]))
      self.assertFalse(sacad.CoverSourceResult.areImageSigsSimilar(img_sig[2], img_sig[1]))

  def test_coverSources(self):
    """ Check all sources return valid results with different parameters. """
    for size in range(300, 1200 + 1, 300):
      for prefer_https in (True, False):
        source_args = (size, 0, prefer_https)
        sources = (sacad.sources.LastFmCoverSource(*source_args),
                   sacad.sources.CoverLibCoverSource(*source_args),
                   sacad.sources.AmazonCdCoverSource(*source_args),
                   sacad.sources.GoogleImagesWebScrapeCoverSource(*source_args),
                   sacad.sources.AmazonDigitalCoverSource(*source_args))
        for source in sources:
          for artist, album in zip(("Michael Jackson", "Björk"), ("Thriller", "Vespertine")):
            results = source.search(album, artist)
            results = sacad.CoverSourceResult.preProcessForComparison(results, size, 0)
            if not (((size > 500) and isinstance(source, sacad.sources.AmazonCdCoverSource)) or
                    ((size > 600) and isinstance(source, sacad.sources.LastFmCoverSource)) or
                    ((size >= 1200) and isinstance(source, sacad.sources.CoverLibCoverSource) and (artist == "Björk"))):
              self.assertGreaterEqual(len(results), 1, "%s %s %s %s %u" % (source.__class__.__name__,
                                                                           " HTTPS" if prefer_https else "",
                                                                           artist,
                                                                           album,
                                                                           size))
            for result in results:
              self.assertTrue(result.urls)
              self.assertIn(result.format, sacad.cover.CoverImageFormat)
              self.assertGreaterEqual(result.size[0], size)


if __name__ == "__main__":
  # logging
  #logging.getLogger().setLevel(logging.DEBUG)
  logging.getLogger().setLevel(logging.CRITICAL + 1)

  # run tests
  unittest.main()



import contextlib
import os
import tempfile


@contextlib.contextmanager
def mkstemp(*args, **kwargs):
  """
  Context manager similar to tempfile.NamedTemporaryFile except the file is not deleted on close, and only the filepath
  is returned
  .. warnings:: Unlike tempfile.mkstemp, this is not secure
  """
  fd, filename = tempfile.mkstemp(*args, **kwargs)
  os.close(fd)
  try:
    yield filename
  finally:
    os.remove(filename)



from .amazoncd import AmazonCdCoverSource, AmazonCdCoverSourceResult
from .amazondigital import AmazonDigitalCoverSource, AmazonDigitalCoverSourceResult
from .lastfm import LastFmCoverSource, LastFmCoverSourceResult
from .google_images import GoogleImagesWebScrapeCoverSource, GoogleImagesCoverSourceResult
from .coverlib import CoverLibCoverSource, CoverLibCoverSourceResult



""" This module provides classes with context managers to help avoid overloading a web API. """

import sqlite3
import time


class ApiAccessRateWatcher:

  def __init__(self, logger, db_filepath=":memory:", min_delay_between_accesses=None, max_accesses_per_day=None):
    self.logger = logger
    self.min_delay_between_accesses = min_delay_between_accesses
    self.max_accesses_per_day = max_accesses_per_day
    self.connexion = sqlite3.connect(db_filepath)
    with self.connexion:
      self.connexion.executescript("""PRAGMA journal_mode = MEMORY;
                                      PRAGMA synchronous = OFF;
                                      CREATE TABLE IF NOT EXISTS access_timestamp (timestamp FLOAT NOT NULL);""")
      self.connexion.execute("DELETE FROM access_timestamp WHERE (strftime('%s', 'now') - timestamp) > 86400;")

  def __enter__(self):
    self.waitAccess()

  def __exit__(self, exc_type, exc_value, traceback):
    self.access()

  def access(self):
    """ Notify the watcher that the API is accessed. """
    with self.connexion:
      self.connexion.execute("INSERT INTO access_timestamp (timestamp) VALUES (?)", (time.time(),))

  def waitAccess(self, timeout=None):
    """ Wait the needed time before sending a request to honor rate limit. Return False if timeout, True otherwise. """
    if self.max_accesses_per_day is not None:
      # daily quota
      now = time.time()
      one_day = 60 * 60 * 24
      one_day_ago = now - one_day
      with self.connexion:
        req_result = self.connexion.execute("""SELECT timestamp
                                               FROM access_timestamp
                                               WHERE timestamp > ?
                                               ORDER BY timestamp DESC
                                               LIMIT 1
                                               OFFSET ?;""",
                                            (one_day_ago,
                                             self.max_accesses_per_day - 1)).fetchone()
      if req_result:
        time_to_wait = req_result[0] - one_day_ago
        if (timeout is not None) and (time_to_wait >= timeout):
          return False
        self.logger.warning("Sleeping for %us because of daily quota" % (time_to_wait))
        time.sleep(time_to_wait)

    if self.min_delay_between_accesses is not None:
      # rate limit
      with self.connexion:
        req_result = self.connexion.execute("""SELECT timestamp
                                               FROM access_timestamp
                                               ORDER BY timestamp DESC
                                               LIMIT 1;""").fetchone()
      if req_result:
        now = time.time()
        last_access_time = req_result[0]
        time_since_last_access = now - last_access_time
        if time_since_last_access < self.min_delay_between_accesses:
          time_to_wait = self.min_delay_between_accesses - time_since_last_access
          if (timeout is not None) and (time_to_wait >= timeout):
            return False
          self.logger.debug("Sleeping for %.2fms because of rate limit" % (time_to_wait * 1000))
          time.sleep(time_to_wait)

    return True



import collections
import operator
import re

import lxml.cssselect
import lxml.etree

from sacad.cover import CoverImageFormat, CoverSourceQuality, CoverSourceResult
from .base import CoverSource


class CoverLibCoverSourceResult(CoverSourceResult):

  def __init__(self, *args, **kwargs):
    super().__init__(*args, source_quality=CoverSourceQuality.NORMAL, **kwargs)


class CoverLibCoverSource(CoverSource):

  """ Cover source that scrapes the coverlib.com site. """

  BASE_URL = "http://coverlib.com/"

  def getSearchUrl(self, album, artist):
    """ See CoverSource.getSearchUrl. """
    url = "%ssearch/" % (__class__.BASE_URL)
    params = collections.OrderedDict()
    params["q"] = "%s %s" % (artist.lower(), album.lower())
    params["Sektion"] = "2"
    return __class__.assembleUrl(url, params)

  def updateHttpHeaders(self, headers):
    """ See CoverSource.updateHttpHeaders. """
    pass

  def parseResults(self, api_data):
    """ See CoverSource.parseResults. """
    results = []

    # parse page
    parser = lxml.etree.HTMLParser()
    html = lxml.etree.XML(api_data.decode("latin-1"), parser)
    album_selector = lxml.cssselect.CSSSelector("div.panel-body div.row")
    cover_selector = lxml.cssselect.CSSSelector("div.col-lg-3")
    info_selector = lxml.cssselect.CSSSelector("p.visible-lg-block.thumbdescr")
    link_selector = lxml.cssselect.CSSSelector("a.dallery-item")
    size_regex = re.compile("Size: ([0-9.]+) x ([0-9.]+)px")

    rank = 1
    for album in album_selector(html):
      try:
        cover = cover_selector(album)[0]  # first cover is front cover
      except IndexError:
        # no results
        break
      desc = info_selector(cover)[0]
      desc_txt = lxml.etree.tostring(desc, encoding="unicode", method="text")

      # get resolution
      re_match = size_regex.search(desc_txt)
      size = tuple(map(int,
                       map(operator.methodcaller("replace", ".", ""),
                           re_match.group(1, 2))))

      # get thumbnail url
      link = link_selector(cover)[0]
      thumbnail_url = link.find("img").get("src")

      # deduce img url without downloading subpage
      cover_id = int(thumbnail_url.rsplit(".", 1)[0].rsplit("/", 1)[1])
      cover_name = link.get("href").rsplit(".", 1)[0].rsplit("/", 1)[1]
      img_url = "%sDownload/%u/%s-Front.JPG" % (__class__.BASE_URL, cover_id, cover_name)

      # assume format is always jpg
      format = CoverImageFormat.JPEG

      # add result
      results.append(CoverLibCoverSourceResult(img_url,
                                               size,
                                               format,
                                               thumbnail_url=thumbnail_url,
                                               rank=rank))
      rank += 1

    return results
